import numpy as np
import matplotlib.pyplot as plt

def distance_matrix(train, test):
    """
    Args:
    train (np.ndarray(np.float32)): La matrice de données d'images d'entraînement de forme (n, m).
    test (np.ndarray(np.float32)): La matrice de données d'images de test de forme (m, p).

    Retourne:
    np.ndarray(np.float32): La matrice des distances euclidiennes L2 entre train et test, de forme (n, p).

    """

    dists = np.sqrt(-2 * np.matmul(train, test.T) + 
                    np.sum(train*train, axis=1, keepdims=True) + 
                    np.sum(test*test, axis=1, keepdims=True).T)
    print('finished calculating dists')

    return dists

def mode(x):
    """
    Args:
    x (float): un tableau de nombres

    Retourne:
    float: le mode de x

    """
    vals, counts = np.unique(x, return_counts=True)

    return vals[np.argmax(counts)]

def knn_predict(dists, labels_train, k):
    """
    Cette fonction prédit les étiquettes pour les exemples de test dans la matrice de distances "dists" en utilisant les k plus proches voisins du test dans l'ensemble d'entraînement.
    Retourne :
    labels_pred (tableau numpy) : tableau de taille (k, num_test) des étiquettes prédites pour les données de test.
    
    """

    indexes_of_knn = np.argsort(dists, axis=0)[0:k, :]
    nearest_labels_pred = labels_train[indexes_of_knn]
    labels_pred = np.array([ mode(label) for label in nearest_labels_pred.T ])

    return labels_pred

def evaluate_knn_for_k(data_train, labels_train, data_test, labels_test, K_max):
    """
    Args:
    data_train (tableau numpy): données d'entraînement
    labels_train (tableau numpy): étiquettes d'entraînement
    data_test (tableau numpy): données de test
    labels_test (tableau numpy): étiquettes de test
    k_max (entier) : nombre maximal de voisins dans le k-NN

    Retourne:
    accuracies (Liste): Liste des précisions des prédictions pour chaque valeur de k.

    """

    accuracies = [0] * K_max
    dists = distance_matrix(data_train, data_test)

    for k in range(1, K_max + 1):
        labels_pred = knn_predict(dists, labels_train, k)
        accuracy = np.sum(labels_pred == labels_test) / len(labels_test)
        accuracies[k - 1] = accuracy

    return accuracies

def plot_accuracy_versus_k(accuracies):
    """
    Cette fonction génère un graphique illustrant la variation de la précision en fonction de k 
    et sauvegarde le graphique dans le dossier /results.
    
    """

    k = len(accuracies)

    fig = plt.figure(figsize=(12, 8))
    plt.plot(np.arange(1, k+1, 1), accuracies, 'o-r')
    plt.title("Variation of the accuracy as a function of k")
    plt.xlabel("k (number of neighbors)")
    plt.ylabel("Accuracy")
    ax = fig.gca()
    ax.set_xticks(np.arange(1, k+1, 1))
    plt.grid(axis='both', which='both')
    plt.savefig(r'C:\Users\hp\Desktop\BE\image-classification\resultats\Knn.png')